@@ -1,5 +1,5 @@
 use serde_json::{json, Value};
-use std::collections::HashMap;
+use std::collections::{BTreeMap, HashMap};
 use uuid::Uuid;
 
 use std::cmp::Ordering;
@@ -119,18 +119,19 @@ impl ProfileBuilder {
                 "processType": 0,
                 "categories": [
                     {
-                        "name": "Idle",
-                        "color": "transparent",
-                    },
-                    {
-                        "name": "Running",
+                        "name": "Regular",
                         "color": "blue",
                     },
                     {
                         "name": "Other",
                         "color": "grey",
                     }
-                ]
+                ],
+                "sampleUnits": {
+                    "time": "ms",
+                    "eventDelay": "ms",
+                    "threadCPUDelta": "µs"
+                }
             },
             "libs": libs,
             "threads": threads,
@@ -147,6 +148,7 @@ pub struct ThreadBuilder {
     start_time: f64,
     end_time: Option<f64>,
     is_main: bool,
+    is_libdispatch_thread: bool,
     stack_table: StackTable,
     frame_table: FrameTable,
     samples: SampleTable,
@@ -154,7 +156,13 @@ pub struct ThreadBuilder {
 }
 
 impl ThreadBuilder {
-    pub fn new(pid: u32, thread_index: u32, start_time: f64, is_main: bool) -> Self {
+    pub fn new(
+        pid: u32,
+        thread_index: u32,
+        start_time: f64,
+        is_main: bool,
+        is_libdispatch_thread: bool,
+    ) -> Self {
         ThreadBuilder {
             pid,
             index: thread_index,
@@ -162,6 +170,7 @@ impl ThreadBuilder {
             start_time,
             end_time: None,
             is_main,
+            is_libdispatch_thread,
             stack_table: StackTable::new(),
             frame_table: FrameTable::new(),
             samples: SampleTable(Vec::new()),
@@ -185,11 +194,26 @@ impl ThreadBuilder {
         self.index
     }
 
-    pub fn add_sample(&mut self, timestamp: f64, frames: &[u64], idle: bool) {
-        let stack_index = self.stack_index_for_frames(frames, idle);
+    pub fn add_sample(&mut self, timestamp: f64, frames: &[u64], cpu_delta: u64) -> Option<usize> {
+        let stack_index = self.stack_index_for_frames(frames);
         self.samples.0.push(Sample {
             timestamp,
             stack_index,
+            cpu_delta,
+        });
+        stack_index
+    }
+
+    pub fn add_sample_same_stack(
+        &mut self,
+        timestamp: f64,
+        previous_stack: Option<usize>,
+        cpu_delta: u64,
+    ) {
+        self.samples.0.push(Sample {
+            timestamp,
+            stack_index: previous_stack,
+            cpu_delta,
         });
     }
 
@@ -197,23 +221,17 @@ impl ThreadBuilder {
         self.end_time = Some(end_time);
     }
 
-    fn stack_index_for_frames(&mut self, frames: &[u64], idle: bool) -> Option<usize> {
+    fn stack_index_for_frames(&mut self, frames: &[u64]) -> Option<usize> {
         let frame_indexes: Vec<_> = frames
             .iter()
-            .enumerate()
-            .map(|(i, &address)| {
-                let frame_idle = idle && i == frames.len() - 1;
-                self.frame_index_for_address(address, frame_idle)
-            })
+            .map(|&address| self.frame_index_for_address(address))
             .collect();
         self.stack_table.index_for_frames(&frame_indexes)
     }
 
-    fn frame_index_for_address(&mut self, address: u64, idle: bool) -> usize {
-        let location_string = format!("0x{:x}", address);
-        let location_string_index = self.string_table.index_for_string(&location_string);
+    fn frame_index_for_address(&mut self, address: u64) -> usize {
         self.frame_table
-            .index_for_frame(location_string_index, idle)
+            .index_for_frame(&mut self.string_table, address)
     }
 
     fn to_json(&self, process_name: &str) -> Value {
@@ -222,6 +240,8 @@ impl ThreadBuilder {
             "GeckoMain".to_string()
         } else if let Some(name) = &self.name {
             name.clone()
+        } else if self.is_libdispatch_thread {
+            "libdispatch".to_string()
         } else {
             format!("Thread <{}>", self.index)
         };
@@ -277,15 +297,18 @@ impl Lib {
 
 #[derive(Debug)]
 struct StackTable {
+    // (parent stack, frame_index)
     stacks: Vec<(Option<usize>, usize)>,
-    index: HashMap<(Option<usize>, usize), usize>,
+
+    // (parent stack, frame_index) -> stack index
+    index: BTreeMap<(Option<usize>, usize), usize>,
 }
 
 impl StackTable {
     pub fn new() -> StackTable {
         StackTable {
             stacks: Vec::new(),
-            index: HashMap::new(),
+            index: BTreeMap::new(),
         }
     }
 
@@ -331,38 +354,39 @@ impl StackTable {
 
 #[derive(Debug)]
 struct FrameTable {
-    frames: Vec<(usize, bool)>,
-    index: HashMap<(usize, bool), usize>,
+    // [string_index]
+    frames: Vec<usize>,
+
+    // address -> frame index
+    index: BTreeMap<u64, usize>,
 }
 
 impl FrameTable {
     pub fn new() -> FrameTable {
         FrameTable {
             frames: Vec::new(),
-            index: HashMap::new(),
+            index: BTreeMap::new(),
         }
     }
 
-    pub fn index_for_frame(&mut self, location_string_index: usize, idle: bool) -> usize {
-        match self.index.get(&(location_string_index, idle)) {
-            Some(frame_index) => *frame_index,
-            None => {
-                let frame_index = self.frames.len();
-                self.frames.push((location_string_index, idle));
-                self.index
-                    .insert((location_string_index, idle), frame_index);
-                frame_index
-            }
-        }
+    pub fn index_for_frame(&mut self, string_table: &mut StringTable, address: u64) -> usize {
+        let frames = &mut self.frames;
+        *self.index.entry(address).or_insert_with(|| {
+            let frame_index = frames.len();
+            let location_string = format!("0x{:x}", address);
+            let location_string_index = string_table.index_for_string(&location_string);
+            frames.push(location_string_index);
+            frame_index
+        })
     }
 
     pub fn to_json(&self) -> Value {
         let data: Vec<Value> = self
             .frames
             .iter()
-            .map(|&(location, idle)| {
-                let category = if idle { 0 } else { 1 };
-                json!([location, false, null, null, null, null, category])
+            .map(|location| {
+                let category = 0;
+                json!([*location, false, null, null, null, null, category])
             })
             .collect();
         json!({
@@ -388,15 +412,14 @@ impl SampleTable {
         let data: Vec<Value> = self
             .0
             .iter()
-            .map(|sample| json!([sample.stack_index, sample.timestamp]))
+            .map(|sample| json!([sample.stack_index, sample.timestamp, 0.0, sample.cpu_delta]))
             .collect();
         json!({
             "schema": {
                 "stack": 0,
                 "time": 1,
-                "responsiveness": 2,
-                "rss": 3,
-                "uss": 4
+                "eventDelay": 2,
+                "threadCPUDelta": 3
             },
             "data": data
         })
@@ -407,6 +430,7 @@ impl SampleTable {
 struct Sample {
     timestamp: f64,
     stack_index: Option<usize>,
+    cpu_delta: u64,
 }
 
 #[derive(Debug)]
