@@ -1,3 +1,5 @@
+use crate::thread_info::time_value;
+
 use super::gecko_profile::ThreadBuilder;
 use mach::mach_types::thread_act_t;
 use std::mem;
@@ -13,7 +15,6 @@ use super::thread_info::{
     thread_basic_info_data_t, thread_extended_info_data_t, thread_identifier_info_data_t,
     thread_info_t, THREAD_BASIC_INFO, THREAD_BASIC_INFO_COUNT, THREAD_EXTENDED_INFO,
     THREAD_EXTENDED_INFO_COUNT, THREAD_IDENTIFIER_INFO, THREAD_IDENTIFIER_INFO_COUNT,
-    TH_STATE_RUNNING,
 };
 
 pub struct ThreadProfiler {
@@ -25,6 +26,8 @@ pub struct ThreadProfiler {
     thread_builder: ThreadBuilder,
     tick_count: usize,
     stack_memory: ForeignMemory,
+    previous_sample_cpu_time: u64,
+    previous_stack: Option<Option<usize>>,
 }
 
 impl ThreadProfiler {
@@ -36,8 +39,8 @@ impl ThreadProfiler {
         now: Instant,
         is_main: bool,
     ) -> kernel_error::Result<Option<Self>> {
-        let tid = match get_thread_id(thread_act) {
-            Ok(tid) => tid as u32,
+        let (tid, is_libdispatch_thread) = match get_thread_id(thread_act) {
+            Ok(info) => info,
             Err(KernelError::MachSendInvalidDest) => return Ok(None),
             Err(KernelError::InvalidArgument) => return Ok(None),
             Err(err) => return Err(err),
@@ -47,6 +50,7 @@ impl ThreadProfiler {
             tid,
             now.duration_since(process_start).as_secs_f64() * 1000.0,
             is_main,
+            is_libdispatch_thread,
         );
         Ok(Some(ThreadProfiler {
             process_start,
@@ -57,6 +61,8 @@ impl ThreadProfiler {
             thread_builder,
             tick_count: 0,
             stack_memory: ForeignMemory::new(task),
+            previous_sample_cpu_time: 0,
+            previous_stack: None,
         }))
     }
 
@@ -80,20 +86,44 @@ impl ThreadProfiler {
             }
         }
 
-        let idle = is_thread_idle(self.thread_act)?;
+        let timestamp = now.duration_since(self.process_start).as_secs_f64() * 1000.0;
+        let cpu_time = get_thread_cpu_time(self.thread_act)?;
+        let cpu_time = cpu_time.0 + cpu_time.1;
+        let cpu_delta = cpu_time.wrapping_sub(self.previous_sample_cpu_time);
 
-        self.stack_scratch_space.clear();
-        get_backtrace(
-            &mut self.stack_memory,
-            self.thread_act,
-            &mut self.stack_scratch_space,
-        )?;
+        if cpu_delta != 0 || self.previous_stack.is_none() {
+            self.stack_scratch_space.clear();
+            get_backtrace(
+                &mut self.stack_memory,
+                self.thread_act,
+                &mut self.stack_scratch_space,
+            )?;
 
-        self.thread_builder.add_sample(
-            now.duration_since(self.process_start).as_secs_f64() * 1000.0,
-            &self.stack_scratch_space,
-            idle,
-        );
+            let stack =
+                self.thread_builder
+                    .add_sample(timestamp, &self.stack_scratch_space, cpu_delta);
+            self.previous_stack = Some(stack);
+        } else if let Some(previous_stack) = self.previous_stack {
+            // No CPU time elapsed since just before the last time we grabbed a stack.
+            // Assume that the thread has done literally zero work and could not have changed
+            // its stack. This considerably reduces the overhead from sampling idle threads.
+            //
+            // More specifically, we hit this path after the following order of events
+            //  - sample n-1:
+            //     - query cpu time, call it A
+            //     - pause the thread
+            //     - walk the stack
+            //     - resume the thread
+            //  - sleep till next sample
+            //  - sample n:
+            //     - query cpu time, notice it is still the same as A
+            //     - add_sample_same_stack with stack from previous sample
+            //
+            self.thread_builder
+                .add_sample_same_stack(timestamp, previous_stack, cpu_delta);
+        }
+
+        self.previous_sample_cpu_time = cpu_time;
 
         Ok(())
     }
@@ -109,7 +139,8 @@ impl ThreadProfiler {
     }
 }
 
-fn get_thread_id(thread_act: thread_act_t) -> kernel_error::Result<u64> {
+/// Returns (tid, is_libdispatch_thread)
+fn get_thread_id(thread_act: thread_act_t) -> kernel_error::Result<(u32, bool)> {
     let mut identifier_info_data: thread_identifier_info_data_t = unsafe { mem::zeroed() };
     let mut count = THREAD_IDENTIFIER_INFO_COUNT;
     unsafe {
@@ -122,7 +153,12 @@ fn get_thread_id(thread_act: thread_act_t) -> kernel_error::Result<u64> {
     }
     .into_result()?;
 
-    Ok(identifier_info_data.thread_id)
+    // This used to check dispatch_qaddr != 0, but it looks like this can happen
+    // even for non-libdispatch threads, for example it happens for rust threads
+    // such as the perfrecord sampler thread.
+    let is_libdispatch_thread = false; // TODO
+
+    Ok((identifier_info_data.thread_id as u32, is_libdispatch_thread))
 }
 
 fn get_thread_name(thread_act: thread_act_t) -> kernel_error::Result<Option<String>> {
@@ -145,7 +181,8 @@ fn get_thread_name(thread_act: thread_act_t) -> kernel_error::Result<Option<Stri
     Ok(if name.is_empty() { None } else { Some(name) })
 }
 
-fn is_thread_idle(thread_act: thread_act_t) -> kernel_error::Result<bool> {
+// (user time, system time) in integer microseconds
+fn get_thread_cpu_time(thread_act: thread_act_t) -> kernel_error::Result<(u64, u64)> {
     // Get the thread name.
     let mut basic_info_data: thread_basic_info_data_t = unsafe { mem::zeroed() };
     let mut count = THREAD_BASIC_INFO_COUNT;
@@ -159,5 +196,12 @@ fn is_thread_idle(thread_act: thread_act_t) -> kernel_error::Result<bool> {
     }
     .into_result()?;
 
-    Ok(basic_info_data.run_state != TH_STATE_RUNNING as i32)
+    Ok((
+        time_value_to_microseconds(&basic_info_data.user_time),
+        time_value_to_microseconds(&basic_info_data.system_time),
+    ))
+}
+
+fn time_value_to_microseconds(tv: &time_value) -> u64 {
+    tv.seconds as u64 * 1_000_000 + tv.microseconds as u64
 }
